#include "Beetle.h"
#include <random>
#include <string>
#include <sstream>
#include <cassert>
#include <cmath>
#include "DebugOutput.h" // Unser Debugging-Header
#include "ConstantData.h"
#include <mmsystem.h>

#define M_PI 3.1415

#pragma comment(lib, "winmm.lib") // fr PlaySound

int Beetle::nextId = 0; // Initialisierung der statischen Variable

Beetle::Beetle() : neuralNetwork(INPUT_SIZE, OUTPUT_SIZE, NUM_HIDDEN_LAYERS, HIDDEN_LAYER_SIZE)
{
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> disX(0, WINDOW_WIDTH - 1);
    std::uniform_int_distribution<> disY(0, WINDOW_HEIGHT - 1 - STATUS_HEIGHT);

    // Zufllige Position fr den Kfer
    x = disX(gen);
    y = disY(gen);

    foodX = FOOD_X;
    foodY = FOOD_Y;
    foodFound = false;
    
    isCurrentGeneration = false;
    id = nextId++;

}

void Beetle::Move()
{   
    // Berechnung der Differenz zwischen der aktuellen Position und der Position des Futters
    double deltaFoodX = GetFoodX() - x;
    double deltaFoodY = GetFoodY() - y;

    // Berechne Winkel zum Futter
    double angleToFood = atan2(deltaFoodY, deltaFoodX);

    // Normalisiere den Winkel im Bereich [0, 1]
    double normalizedAngle = (angleToFood + M_PI) / (2 * M_PI);

    double moveAngle = 0; // Bewegungsrichtung 

    if (TEST_WITHOUT_NN)
    {
        moveAngle = normalizedAngle * 2 * M_PI - M_PI; // Normalisiere den Winkel im Bereich [-PI, PI]
    }
    else // NN wird verwendet
    {
        // Der normalisierte Winkel ist der Input fr das NN
        std::vector<double> inputs;
        inputs.push_back(normalizedAngle);
    
        // Der Output des NN bestimmt die Bewegungsrichtung
        std::vector<double> outputs = neuralNetwork.FeedForward(inputs);
        moveAngle = outputs[0] * 2 * M_PI - M_PI; // Normalisiere den Winkel im Bereich [-PI, PI]
    }

    // Ermittle deltaX und deltaY aus der Bewegungsrichtung
    double deltaX = cos(moveAngle) * SPEED;
    double deltaY = sin(moveAngle) * SPEED;

    // Bestimme die neue Position
    double newX = x + deltaX;
    if (newX >= WINDOW_WIDTH) newX = WINDOW_WIDTH - 1;
    if (newX < 0) newX = 0;
    double newY = y + deltaY;
    if (newY >= WINDOW_HEIGHT - STATUS_HEIGHT) newY = WINDOW_HEIGHT - STATUS_HEIGHT - 1;
    if (newY < 0) newY = 0;

    // Setze neue Koordinaten des Kfers
    x = static_cast<int>(newX);
    y = static_cast<int>(newY);

    if (((x <= (foodX + FOOD_HIT) && x >= (foodX - FOOD_HIT))) && ((y <= (foodY + FOOD_HIT) && y >= (foodY - FOOD_HIT))) && !foodFound)
    {
        foodFound = true;
        TotalHitsPerGeneration++;
        if (!::PlaySound(TEXT("food_sound.wav"), NULL, SND_FILENAME | SND_ASYNC))
        {
            std::ostringstream oss;
            oss << "\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>Error playing sound";
            DebugOutput(oss.str());
        }
    }

    // Fge neues Segment hinzu
    TrailSegment segment = { x, y, 0 };
    trail.push_back(segment);

    // ltestes Segment entfernen, wenn die Spur-Lnge berschritten wird
    if (!trail.empty() && (trail.size() > TRAIL_MAX_LENGTH))
    {
        TrailSegment oldSegment = trail.front();
        trail.erase(trail.begin());
    }

    // Altern der Spur erhhen
    for (auto& segment : trail)
    {
        segment.age++;
    }
}



void Beetle::Draw(HDC hdc)
{
    // Farben definieren
    COLORREF foodColor = BLUE;
    COLORREF OldFoodColor = GRAY;
    COLORREF currentBeetleColor = RED;
    COLORREF oldBeetleColor = LIGHTGRAY;

    // Futter zeichnen (als gefllter Kreis)
    HBRUSH hFoodBrush = CreateSolidBrush(foodColor);
    HBRUSH hOldFoodBrush = CreateSolidBrush(OldFoodColor);

    if (isCurrentGeneration)
    {
        SelectObject(hdc, hFoodBrush);
    }
    else
    {
        SelectObject(hdc, hOldFoodBrush);
    }

    // Futter zeichnen (als gefllter Kreis)
    Ellipse(hdc, GetFoodX() - FOOD_SIZE, GetFoodY() - FOOD_SIZE, GetFoodX() + FOOD_SIZE, GetFoodY() + FOOD_SIZE);

    DeleteObject(hFoodBrush);
    DeleteObject(hOldFoodBrush);

    // Spur zeichnen
    for (auto& segment : trail)
    {
        // Aktualisiere das Alter des Segments und passe die Transparenz an
        int alpha = 0 + (segment.age * FADING_FACTOR); // Schnelleres Verblassen durch hheren Faktor
        if (alpha > 255) alpha = 255;

        // Transparente Farbe berechnen (Gelb nach Wei)
        int red = TRAIL_COLOR_PARAMETER1;
        int green = 255;
        int blue = (255 * alpha) / 255;
        COLORREF trailColor = RGB(red, green, blue);

        // Erstelle einen Pinsel mit der berechneten Farbe
        HBRUSH hTrailBrush = CreateSolidBrush(trailColor);
        HBRUSH hOldTrailBrush = (HBRUSH)SelectObject(hdc, hTrailBrush);

        // Erstelle einen Stift mit der gleichen Farbe
        HPEN hTrailPen = CreatePen(PS_SOLID, 1, trailColor);
        HPEN hOldPen = (HPEN)SelectObject(hdc, hTrailPen);

        // Zeichnen des Segments
        Ellipse(hdc, segment.x - BEETLE_SIZE, segment.y - BEETLE_SIZE, segment.x + BEETLE_SIZE, segment.y + BEETLE_SIZE);

        // Selektiere den alten Pinsel und Stift zurck
        SelectObject(hdc, hOldTrailBrush);
        SelectObject(hdc, hOldPen);

        DeleteObject(hTrailBrush);
        DeleteObject(hTrailPen);
    }

    // Pinsel fr die aktuellen und alten Kfer erstellen
    HPEN hCurrentBeetlePen = CreatePen(PS_SOLID, 1, currentBeetleColor);
    HBRUSH hCurrentBeetleBrush = CreateSolidBrush(currentBeetleColor);
    HPEN hOldBeetlePen = CreatePen(PS_SOLID, 1, oldBeetleColor);
    HBRUSH hOldBeetleBrush = CreateSolidBrush(oldBeetleColor);

    if (!isCurrentGeneration)
    {
        HPEN hOldPen = (HPEN)SelectObject(hdc, hOldBeetlePen);
        HBRUSH hOldBrush = (HBRUSH)SelectObject(hdc, hOldBeetleBrush);
        // Kfer zeichnen (als gefllter Kreis)
        Ellipse(hdc, x - BEETLE_SIZE, y - BEETLE_SIZE, x + BEETLE_SIZE, y + BEETLE_SIZE);
        SelectObject(hdc, hOldPen);
        SelectObject(hdc, hOldBrush);        
    }
    else
    {
        HPEN hOldPen = (HPEN)SelectObject(hdc, hCurrentBeetlePen);
        HBRUSH hOldBrush = (HBRUSH)SelectObject(hdc, hCurrentBeetleBrush);
        // Kfer zeichnen (als gefllter Kreis)
        Ellipse(hdc, x - BEETLE_SIZE, y - BEETLE_SIZE, x + BEETLE_SIZE, y + BEETLE_SIZE);
        SelectObject(hdc, hOldPen);
        SelectObject(hdc, hOldBrush);
    }

    DeleteObject(hCurrentBeetlePen);
    DeleteObject(hCurrentBeetleBrush);
    DeleteObject(hOldBeetlePen);
    DeleteObject(hOldBeetleBrush);
}


Beetle Beetle::Crossover(const Beetle& other, double crossoverRate, int type) const
{
    Beetle offspring;

    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_real_distribution<> dis(0.0, 1.0);
    std::uniform_int_distribution<> pointDist(0, static_cast<int>(this->GetNeuralNetwork().GetWeights().size() - 1));

    std::vector<double> offspringWeights = offspring.GetNeuralNetwork().GetWeights();
    const std::vector<double>& thisWeights = this->GetNeuralNetwork().GetWeights();
    const std::vector<double>& otherWeights = other.GetNeuralNetwork().GetWeights();

    // Gewichtsvektoren mssen dieselbe Gre haben
    if (thisWeights.size() != otherWeights.size())
    {
        throw std::runtime_error("Crossover: Weight vectors must be of the same size");
    }

    /*
    std::ostringstream oss;
    oss << "\nthisWeights.size: " << thisWeights.size();
    DebugOutput(oss.str());
    */

    size_t crossoverPoint = 0;
    size_t point1 = 0, point2 = 0;

    switch (type)
    {
    case 0: // Uniforme Kreuzung
        // Fr jedes Gewicht entscheidet ein Zufallswert (dis(gen)), 
        // ob das Gewicht vom aktuellen Kfer (thisWeights[i]) 
        // oder vom anderen Elternteil (otherWeights[i]) an den Nachkommen weitergegeben wird. 
        // Dies basiert auf der crossoverRate.
        for (size_t i = 0; i < thisWeights.size(); ++i)
        {
            offspringWeights[i] = (dis(gen) < crossoverRate) ? thisWeights[i] : otherWeights[i];
        }
        break;

    case 1: // Einpunkt-Kreuzung
        // Ein einziger zuflliger Kreuzungspunkt wird ausgewhlt (crossoverPoint), 
        // wobei der pointDist-Verteiler verwendet wird, um die Position des Kreuzungspunktes zu bestimmen.
        // Alle Gewichte vor diesem Punkt stammen vom aktuellen Kfer (thisWeights), 
        // und alle Gewichte ab diesem Punkt stammen vom anderen Elternteil (otherWeights).

        crossoverPoint = pointDist(gen); // Der pointDist-Verteiler wird verwendet, 
        // um die Kreuzungspunkte zufllig auszuwhlen.

        for (size_t i = 0; i < thisWeights.size(); ++i)
        {
            offspringWeights[i] = (i < crossoverPoint) ? thisWeights[i] : otherWeights[i];
        }
        break;

    case 2: // Zweipunkt-Kreuzung
        // Zwei zufllige Kreuzungspunkte werden ausgewhlt (point1 und point2), 
        // wobei der pointDist-Verteiler verwendet wird, um die Position der Kreuzungspunkte zu bestimmen. 
        // Wenn der erste Punkt grer ist als der zweite, werden die Punkte vertauscht, 
        // um sicherzustellen, dass der erste Punkt kleiner oder gleich dem zweiten ist (std::swap).
        // Gewichte vor dem ersten Punkt und nach dem zweiten Punkt stammen vom aktuellen Kfer (thisWeights), 
        // whrend die Gewichte zwischen den Punkten vom anderen Elternteil (otherWeights) stammen.

        point1 = pointDist(gen); // Der pointDist-Verteiler: zuflliger erster Kreuzungspunkt
        point2 = pointDist(gen); // Der pointDist-Verteiler: zuflliger zweiter Kreuzungspunkt

        if (point1 > point2) std::swap(point1, point2); // Die std::swap-Funktion stellt sicher, 
        // dass der erste Kreuzungspunkt kleiner oder gleich dem zweiten ist, 
        // was die Implementierung der Zweipunkt-Kreuzung vereinfacht.

        for (size_t i = 0; i < thisWeights.size(); ++i)
        {
            offspringWeights[i] = (i < point1 || i > point2) ? thisWeights[i] : otherWeights[i];
        }
        break;

    default:
        throw std::invalid_argument("Invalid crossover type");
    }

    offspring.GetNeuralNetwork().SetWeights(offspringWeights);
    offspring.foodFound = false;
    return offspring;
}


void Beetle::Mutate(double mutationRate)
{
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_real_distribution<> dis(0.0, 1.0);
    std::normal_distribution<> mutationDist(0.0, 1.0);

    std::vector<double> weights = neuralNetwork.GetWeights();

    for (auto& weight : weights)
    {
        // Wenn die Zufallszahl kleiner als die mutationRate ist, 
        // wird das Gewicht um einen Wert mutationDist(gen) gendert, 
        // welcher aus der Normalverteilung stammt.
        // Diese Kombination aus einer uniformen Verteilung zur Entscheidungsfindung 
        // und einer Normalverteilung zur Bestimmung der Mutationsgre ist eine bliche Vorgehensweise in genetischen Algorithmen und evolutiven Strategien, 
        // da sie eine gute Balance zwischen Exploration (durch Mutationen) 
        // und Exploitation (durch Kreuzung) ermglicht.
        if (dis(gen) < mutationRate)
        {
            weight += mutationDist(gen);
        }
    }

    neuralNetwork.SetWeights(weights);
    foodFound = false;
}

NeuralNetwork& Beetle::GetNeuralNetwork() {
    return neuralNetwork;
}

const NeuralNetwork& Beetle::GetNeuralNetwork() const {
    return neuralNetwork;
}

