
#define NOMINMAX
#include "GeneticAlgorithm.h"
#include <algorithm>
#include <random>
#include <cassert>
#include <sstream>
#include "DebugOutput.h"
#include "Beetle.h"
#include "NeuralNetwork.h"
#include "ConstantData.h"

double GeneticAlgorithm::CalculateAverageFitness(const std::vector<Beetle>& beetles) {
    double totalFitness = 0.0;
    for (const auto& beetle : beetles) {
        double distance = sqrt(pow(beetle.GetX() - FOOD_X, 2) + pow(beetle.GetY() - FOOD_Y, 2));
        double fitness = 1.0 / (distance + 1); // Vermeiden von Division durch Null
        totalFitness += fitness;
    }
    return totalFitness / beetles.size();
}

double GeneticAlgorithm::GetMutationRate(double previousBestFitness, double currentBestFitness) {
    static double mutationRate = 0.01; // Startwert
    double fitnessImprovement = currentBestFitness - previousBestFitness;
    if (fitnessImprovement < 0.01) {
        mutationRate = std::min(0.1, mutationRate + 0.005); // Erhht die Mutationsrate, bis zu einem Maximum von 10%
    }
    else {
        mutationRate = std::max(0.001, mutationRate - 0.001); // Verringert die Mutationsrate, bis zu einem Minimum von 0.1%
    }
    return mutationRate;
}

double GeneticAlgorithm::GetCrossoverRate(double previousAverageFitness, double currentAverageFitness) {
    static double crossoverRate = 0.7; // Startwert
    double fitnessImprovement = currentAverageFitness - previousAverageFitness;
    if (fitnessImprovement < 0.01) {
        crossoverRate = std::min(0.9, crossoverRate + 0.01); // Erhht die Kreuzungsrate, bis zu einem Maximum von 90%
    }
    else {
        crossoverRate = std::max(0.5, crossoverRate - 0.01); // Verringert die Kreuzungsrate, bis zu einem Minimum von 50%
    }
    return crossoverRate;
}

double CalculateFitness(const Beetle& beetle) {
    double distance = sqrt(pow(beetle.GetX() - FOOD_X, 2) + pow(beetle.GetY() - FOOD_Y, 2));
    return 1.0 / (distance + 1); // Vermeiden von Division durch Null
}

std::pair<std::vector<Beetle>, std::pair<int, int>> GeneticAlgorithm::NextGeneration(const std::vector<Beetle>& currentGeneration, double mutationRate, double crossoverRate)
{
    
    std::vector<Beetle> newGeneration;
    newGeneration.reserve(currentGeneration.size());
    int crossoverCount = 0;
    int cloneCount = 0;

    // Fitness-Berechnung
    std::vector<std::pair<Beetle, double>> fitnessScores;
    for (const auto& beetle : currentGeneration) {
        double fitness = CalculateFitness(beetle);
        fitnessScores.push_back({ beetle, fitness });
    }

    // Sortiere nach Fitness
    std::sort(fitnessScores.begin(), fitnessScores.end(), [](const auto& a, const auto& b) {
        return a.second > b.second;
        });

    // Selektion: Whle die besten Kfer fr die Reproduktion
    int numParents = static_cast<int>(currentGeneration.size() / BEST_SELECTOR);
    numParents = std::max(1, numParents); // Mindestens ein Elternteil

    std::vector<Beetle> parents;
    for (int i = 0; i < numParents; ++i) {
        parents.push_back(fitnessScores[i].first);
    }

    // Kreuzung und Mutation fr die neue Generation
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> dist(0, numParents - 1);

    while (newGeneration.size() < currentGeneration.size()) {
        Beetle parent1 = parents[dist(gen)];
        Beetle parent2 = parents[dist(gen)];

        if (CLONING_FORBIDDEN)
        {
            // Vermeide das Kreuzen des gleichen Elternteils
            while (parent1.GetID() == parent2.GetID()) {
                parent2 = parents[dist(gen)];
            }
        }

        Beetle offspring;
        if (parent1.GetID() == parent2.GetID()) {
            offspring = parent1; // Klonen
            cloneCount++;
        }
        else {
            offspring = parent1.Crossover(parent2, crossoverRate, CROSSOVER_TYPE); // Kreuzung
            crossoverCount++;
        }
        offspring.Mutate(mutationRate);
        offspring.SetFoodFound(false);
        newGeneration.push_back(offspring);
    }

    return { newGeneration, { crossoverCount, cloneCount } };
}

