#define NOMINMAX
#include <windows.h>
#include <limits>
#include <cassert>
#include <vector>
#include <algorithm>
#include <random>
#include <string>
#include <sstream>
#include <cmath>
#include <chrono>
#include <commctrl.h> // Fr die Statusleiste
#include "Beetle.h"
#include "GeneticAlgorithm.h"
#include "DebugOutput.h"
#include "ConstantData.h"
#include <mmsystem.h>

#pragma comment(lib, "winmm.lib") // fr PlaySound
#pragma comment(lib, "comctl32.lib") 
#pragma comment(lib, "Msimg32.lib")

std::ostringstream oss; //Debug

int FOOD_X = WINDOW_WIDTH / 2;
int FOOD_Y = WINDOW_HEIGHT / 2;
int generationStep = 0;
NeuralNetwork bestNN (INPUT_SIZE, OUTPUT_SIZE, NUM_HIDDEN_LAYERS, HIDDEN_LAYER_SIZE);
int currentGeneration = 0;
int TotalHitsPerGeneration = 0;
double bestFitness = 0.0;
double previousBestFitness = 0.0;
double previousAverageFitness = 0.0;
std::vector<Beetle> currentBeetles(POPULATION_SIZE);
std::vector<Beetle> previousBeetles;
std::pair<int, int> counts;

HWND hStatus;

GeneticAlgorithm geneticAlgorithm; // Instanz von GeneticAlgorithm

void ClearScreen(HWND hwnd)
{
    // Handle des Gertekontexts (device context) des Fensters erhalten
    HDC hdc = GetDC(hwnd);

    // Clientbereich des Fensters ermitteln
    RECT clientRect;
    GetClientRect(hwnd, &clientRect);

    // Farbe Wei erstellen
    HBRUSH hWhiteBrush = CreateSolidBrush(WHITE);

    // Clientbereich mit Wei fllen
    FillRect(hdc, &clientRect, hWhiteBrush);

    // Freigeben der erstellten Ressourcen
    DeleteObject(hWhiteBrush);
    ReleaseDC(hwnd, hdc);
}

void PlayGenerationSound() {
    ::PlaySound(TEXT("generation_sound.wav"), NULL, SND_FILENAME | SND_ASYNC);
}

static void UpdateStatusBar(HWND hwnd, int generation, double bestFitness, int step, size_t populationSize, double avgFitness, double mutationRate, double crossoverRate, size_t elapsedTime, int bestBeetleX, int bestBeetleY, int crossoverCount, int cloneCount, int TotalHitsPerGeneration) 
{
    wchar_t statusText[512];
    swprintf_s(statusText, 
    L"Generation: %d | Best Fitness: %.4f | Step: %d | Pop. Size: %zu | Avg Fitness: %.4f | Mutation: %.2f%% | Crossover: %.2f%% | Time: %zus | Best Beetle (x,y): (%d, %d) | Crossover: %d | Clones: %d | Total Hits: %d (%.1f%%)",
        generation+1, bestFitness, step+1, populationSize, avgFitness, mutationRate * 100, crossoverRate * 100, elapsedTime, bestBeetleX, bestBeetleY, crossoverCount, cloneCount, TotalHitsPerGeneration, (100.0*TotalHitsPerGeneration)/POPULATION_SIZE);

    SendMessage(hwnd, SB_SETTEXT, 0, (LPARAM)statusText);

    // Nur den Bereich der Statusleiste invalidieren
    RECT rect;
    SendMessage(hwnd, SB_GETRECT, 0, (LPARAM)&rect);
    InvalidateRect(hwnd, &rect, TRUE);
}

size_t GetElapsedTimeInSeconds() {
    static auto startTime = std::chrono::steady_clock::now();
    auto elapsed = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::steady_clock::now() - startTime);
    return elapsed.count();
}

int GetBestBeetleX(const std::vector<Beetle>& beetles) {
    double bestFitness = 0.0;
    int bestX = 0;
    for (const auto& beetle : beetles) {
        double distance = sqrt(pow(beetle.GetX() - FOOD_X, 2) + pow(beetle.GetY() - FOOD_Y, 2));
        double fitness = 1.0 / (distance + 1);
        if (fitness > bestFitness) {
            bestFitness = fitness;
            bestX = beetle.GetX();
        }
    }
    return bestX;
}

int GetBestBeetleY(const std::vector<Beetle>& beetles) {
    double bestFitness = 0.0;
    int bestY = 0;
    for (const auto& beetle : beetles) {
        double distance = sqrt(pow(beetle.GetX() - FOOD_X, 2) + pow(beetle.GetY() - FOOD_Y, 2));
        double fitness = 1.0 / (distance + 1);
        if (fitness > bestFitness) {
            bestFitness = fitness;
            bestY = beetle.GetY();
        }
    }
    return bestY;
}

void DrawWeightHeatmap(HDC hdc, const std::vector<double>& weights, int x, int y, int cellSize)
{
    size_t num_weights = weights.size();
    int input_to_hidden_size = INPUT_SIZE * HIDDEN_LAYER_SIZE;
    int hidden_to_hidden_size = HIDDEN_LAYER_SIZE * HIDDEN_LAYER_SIZE;
    int hidden_to_output_size = HIDDEN_LAYER_SIZE * OUTPUT_SIZE;

    double min_weight = std::numeric_limits<double>::max();
    double max_weight = std::numeric_limits<double>::lowest();

    // Berechnung von min_weight und max_weight
    for (const double& weight : weights) {
        if (weight < min_weight) min_weight = weight;
        if (weight > max_weight) max_weight = weight;
    }

    // Verhindert Division durch Null, falls min_weight == max_weight
    if (max_weight == min_weight) {
        max_weight = min_weight + 1.0;
    }

    // Schriftart einstellen
    HFONT hFont = CreateFont(cellSize - 5, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, DEFAULT_CHARSET,
        OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
        FIXED_PITCH | FF_MODERN, L"Consolas");
    HFONT hOldFont = (HFONT)SelectObject(hdc, hFont);

    auto draw_rectangle = [&](int row, int col, int color_intensity) 
    {
        COLORREF color = RGB(color_intensity, 0, 255 - color_intensity);
        HBRUSH brush = CreateSolidBrush(color);
        HBRUSH oldBrush = (HBRUSH)SelectObject(hdc, brush);

        Rectangle(hdc, x + col * cellSize, y + row * cellSize, x + (col + 1) * cellSize, y + (row + 1) * cellSize);

        SelectObject(hdc, oldBrush);
        DeleteObject(brush);
    };

    auto draw_text = [&](const std::string& text, int row_offset) 
    {
        TextOutA(hdc, x, y + row_offset * cellSize - cellSize / 2, text.c_str(), static_cast<int>(text.length()));
    };

    int row = 0;
    int col = 0;

    // Zeichnen der Gewichte zwischen Input und erstem Hidden Layer
    ++row;// Bild oben verdeckt durch Caption
    draw_text("Input --> HiddenLayer1", row);
    ++row;
    for (int i = 0; i < input_to_hidden_size; ++i) {
        double weight = weights[i];
        int color_intensity = static_cast<int>((weight - min_weight) / (max_weight - min_weight) * 255);
        draw_rectangle(row, col, color_intensity);

        if (++col >= HIDDEN_LAYER_SIZE) {
            col = 0;
            ++row;
        }
    }

    // Leerzeile zwischen den Layern
    ++row;

    // Zeichnen der Gewichte zwischen den Hidden Layers
    for (int l = 0; l < (NUM_HIDDEN_LAYERS-1); ++l) {
        std::ostringstream oss;
        oss << "HiddenLayer" << l + 1 << " --> HiddenLayer" << l + 2;
        draw_text(oss.str(), row);
        ++row;
        for (int i = 0; i < hidden_to_hidden_size; ++i) {
            int index = input_to_hidden_size + l * hidden_to_hidden_size + i;
            double weight = weights[index];
            int color_intensity = static_cast<int>((weight - min_weight) / (max_weight - min_weight) * 255);
            draw_rectangle(row, col, color_intensity);

            if (++col >= HIDDEN_LAYER_SIZE) {
                col = 0;
                ++row;
            }
        }
        // Leerzeile zwischen den Layern
        ++row;
    }

    // Zeichnen der Gewichte zwischen letztem Hidden Layer und Output Layer
    std::ostringstream oss;
    oss << "HiddenLayer" << NUM_HIDDEN_LAYERS << " --> Output";
    draw_text(oss.str(), row);    
    ++row;
    for (int i = 0; i < hidden_to_output_size; ++i) {
        int index = input_to_hidden_size + (NUM_HIDDEN_LAYERS-1) * hidden_to_hidden_size + i;
        double weight = weights[index];
        int color_intensity = static_cast<int>((weight - min_weight) / (max_weight - min_weight) * 255);
        draw_rectangle(row, col, color_intensity);

        if (++col >= HIDDEN_LAYER_SIZE) {
            col = 0;
            ++row;
        }
    }

    // Originale Schriftart wiederherstellen
    SelectObject(hdc, hOldFont);
    DeleteObject(hFont);
}


// Fensterprozedur
LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) 
{
    switch (uMsg) 
    {
        case WM_DESTROY:
            PostQuitMessage(0);
            return 0;
        case WM_PAINT: 
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd, &ps);
            // Erstellen eines kompatiblen DC und Bitmap
            HDC hdcMem = CreateCompatibleDC(hdc);
            HBITMAP hbmMem = CreateCompatibleBitmap(hdc, WINDOW_WIDTH + WINDOW_EXTRAWIDTH, WINDOW_HEIGHT);
            HGDIOBJ hOld = SelectObject(hdcMem, hbmMem);
            
            // Hintergrund lschen
            HBRUSH hbrBkGnd = CreateSolidBrush(RGB(255, 255, 255)); // Weier Hintergrund
            FillRect(hdcMem, &ps.rcPaint, hbrBkGnd);
            DeleteObject(hbrBkGnd);

            // bestes NN darstellen
            // Gewichte als Heatmap visualisieren
            std::vector<double> weights = bestNN.GetWeights();
            DrawWeightHeatmap(hdcMem, weights, WINDOW_WIDTH+10, 0, CELL_SIZE);

            // Zeichne vorherige Generation in grau
            for (auto& beetle : previousBeetles) {
                beetle.SetCurrentGeneration(false);
                beetle.Draw(hdcMem);
            }

            // Zeichne aktuelle Generation in rot
            for (auto& beetle : currentBeetles) {
                beetle.SetCurrentGeneration(true);
                beetle.Draw(hdcMem);
            }
            
            // Kopiere den Speicher-DC in den Bildschirm-DC
            BitBlt(hdc, 0, 0, WINDOW_WIDTH + WINDOW_EXTRAWIDTH, WINDOW_HEIGHT, hdcMem, 0, 0, SRCCOPY);

            // Aufrumen
            SelectObject(hdcMem, hOld);
            DeleteObject(hbmMem);
            DeleteDC(hdcMem);

            EndPaint(hwnd, &ps);
            return 0;
        }
        case WM_TIMER:
        {
            UINT_PTR timerID = (UINT_PTR)wParam;

            switch (timerID)
            {
            case 1:
                // Statusleiste aktualisieren
            UpdateStatusBar(hStatus, currentGeneration, bestFitness, generationStep, currentBeetles.size(),
                geneticAlgorithm.CalculateAverageFitness(currentBeetles), geneticAlgorithm.GetMutationRate(previousBestFitness, bestFitness),
                geneticAlgorithm.GetCrossoverRate(previousAverageFitness, geneticAlgorithm.CalculateAverageFitness(currentBeetles)),
                GetElapsedTimeInSeconds(), GetBestBeetleX(currentBeetles), GetBestBeetleY(currentBeetles), counts.first, counts.second, TotalHitsPerGeneration);

            // Bewegungslogik fr die Kfer
            for (auto& beetle : currentBeetles) { beetle.Move(); }

            // Generationsfortschritt zhlen und Generationen aktualisieren
            if (++generationStep >= GENERATION_INTERVAL)
            {
                oss << "\ncurrentGeneration: " << currentGeneration
                    << " generationStep: " << generationStep;
                DebugOutput(oss.str());

                previousBestFitness = bestFitness;
                previousAverageFitness = geneticAlgorithm.CalculateAverageFitness(previousBeetles);
                generationStep = 0;
                currentGeneration++;
                TotalHitsPerGeneration = 0;
                
                // Setze alle Kfer in der aktuellen Generation auf false, bevor sie zur alten Generation werden
                for (auto& beetle : currentBeetles) {
                    beetle.SetCurrentGeneration(false);
                }

                // Speichere die aktuelle Generation in previousBeetles
                previousBeetles = currentBeetles;

                oss << "\nbeetle size old: " << previousBeetles.size();
                DebugOutput(oss.str());

                // Neue Position fr das Futter
                std::random_device rd;
                std::mt19937 gen(rd());
                std::uniform_int_distribution<> disX(WINDOW_WIDTH / 2  - FOOD_SHIFT_X, WINDOW_WIDTH / 2  + FOOD_SHIFT_X); // Bereich fr x-Koordinate des Futters
                std::uniform_int_distribution<> disY(WINDOW_HEIGHT / 2 - FOOD_SHIFT_Y, WINDOW_HEIGHT / 2 + FOOD_SHIFT_Y); // Bereich fr y-Koordinate des Futters

                // Zufllige Position fr das Futter im definierten Bereich
                FOOD_X = disX(gen);
                FOOD_Y = disY(gen);

                // Neue Generation erzeugen
                auto result = geneticAlgorithm.NextGeneration(previousBeetles, geneticAlgorithm.GetMutationRate(previousBestFitness, bestFitness),
                    geneticAlgorithm.GetCrossoverRate(previousAverageFitness, geneticAlgorithm.CalculateAverageFitness(previousBeetles)));
                std::vector<Beetle> newGeneration = result.first;
                counts = result.second;

                oss << "\nbeetle size new: " << currentBeetles.size();
                DebugOutput(oss.str());

                // Setze alle Kfer in der neuen Generation auf true
                for (auto& beetle : currentBeetles) {
                    beetle.SetCurrentGeneration(true);
                    beetle.SetFoodX(FOOD_X);
                    beetle.SetFoodY(FOOD_Y);
                    beetle.SetFoodFound(false);
                }

                ::PlayGenerationSound(); // Sound abspielen, wenn eine neue Generation erscheint

                // Beste Fitness der neuen Generation berechnen
                bestFitness = 0.0;
                for (const auto& beetle : currentBeetles) 
                {
                    double distance = sqrt(pow(beetle.GetX() - FOOD_X, 2) + pow(beetle.GetY() - FOOD_Y, 2));
                    double fitness = 1.0 / (distance + 1); // Vermeiden von Division durch Null
                    if (fitness > bestFitness) {
                        bestFitness = fitness;
                        bestNN = beetle.GetNeuralNetwork();
                    }
                }
                // Zeichnen der Kfer nach Generationswechsel
                InvalidateRect(hwnd, NULL, TRUE);
            }
            break;

            case 2:
                ClearScreen(hwnd); //Artefakte beseitigen
            break;

            default:
                // Handler fr unbekannte Timer-IDs
            break;
            }   
        }
        
        // Fenster aktualisieren
        InvalidateRect(hwnd, NULL, TRUE);
        return 0;
    }    
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow) {
    INITCOMMONCONTROLSEX icex; // Initialisieren der Common Controls
    icex.dwSize = sizeof(INITCOMMONCONTROLSEX);
    icex.dwICC = ICC_WIN95_CLASSES; // Korrekte Initialisierung der Common Controls
    InitCommonControlsEx(&icex);

    constexpr wchar_t CLASS_NAME[] = L"BeetleWindowClass";
    WNDCLASS wc = {};
    wc.lpfnWndProc = WindowProc;
    wc.hInstance = hInstance;
    wc.lpszClassName = CLASS_NAME;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);

    RegisterClass(&wc);

    HWND hwnd = CreateWindowEx(
        0,
        CLASS_NAME,
        L"Beetles with Neural Network search Food - Improvement by Genetic Algorithm",
        WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
        CW_USEDEFAULT, CW_USEDEFAULT, WINDOW_WIDTH + WINDOW_EXTRAWIDTH, WINDOW_HEIGHT,
        NULL,
        NULL,
        hInstance,
        NULL
    );

    if (hwnd == NULL) {
        return 0;
    }

    // Erstellen der Statusleiste
    hStatus = CreateWindowEx(
        0, STATUSCLASSNAME, NULL,
        WS_CHILD | WS_VISIBLE, 
        0, 0, 0, 0,
        hwnd, (HMENU)1, GetModuleHandle(NULL), NULL
    );

    ShowWindow(hwnd, nCmdShow);
    UpdateWindow(hwnd); // UpdateWindow aufrufen, um sicherzustellen, dass das Fenster gezeichnet wird

    // Timer 
    SetTimer(hwnd, 1, STATUS_UPDATE_INTERVAL1, NULL);  // <--- Wichtig, hier hwnd
    //SetTimer(hwnd, 2, STATUS_UPDATE_INTERVAL2, NULL);  // <--- Wichtig, hier hwnd

    MSG msg = {};
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);

        // Eine kurze Verzgerung hinzufgen
        Sleep(SLEEP_TIME);
    }

    return 0;
}


